// Forces calculation test 1    08/01/2016

using System;
using System.Diagnostics;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

using BEPUphysics;
using BEPUutilities.Threading;
using BEPUphysicsDrawer.Models;
using BEPUphysics.Entities.Prefabs;
using BEPUphysics.Entities;
using BEPUphysics.PositionUpdating;
using BEPUphysics.CollisionShapes.ConvexShapes;

using BEPUphysics.Constraints.SolverGroups;
using BEPUphysics.CollisionShapes;
using System.Collections.Generic;
using BEPUphysics.Settings;
using BEPUphysics.Constraints;
using ConversionHelper;


namespace ForcesOne
{
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        SpriteFont osdfont;

        public ModelDrawer ModelDrawer;

        public KeyboardState KeyboardInput;
        public KeyboardState PreviousKeyboardInput;
        public MouseState MouseInput;
        public MouseState PreviousMouseInput;

        public Space Space;

        Matrix world;
        Matrix view;
        Matrix projection;

        Entity ground;
        Entity demo;

        bool levitateMode = false;
        bool pause = false;

        BEPUutilities.Vector3 LinearMomentum, PreviousLinearMomentum;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            Space = new Space();

            // BEPU fine tuning
            Space.Solver.IterationLimit = 100;
            //CollisionResponseSettings.Softness = 0f;
            CollisionDetectionSettings.AllowedPenetration = 0.002f;
            //SolverSettings.DefaultMinimumImpulse = 0f;
            //BEPUutilities.Toolbox.Epsilon = 0.002f;
            //BEPUutilities.Toolbox.BigEpsilon = 0.002f;
            //InertiaHelper.InertiaTensorScale = 1.0f;

            graphics.PreferredBackBufferWidth = 1280;
            graphics.PreferredBackBufferHeight = 720;

            Space.ForceUpdater.Gravity = new BEPUutilities.Vector3(0, -9.81f, 0f);
        }


        protected override void Initialize()
        {
            ModelDrawer = new InstancedModelDrawer(this);
            ModelDrawer.Clear();

            world = Matrix.CreateTranslation(new Vector3(0, 0, 0));
            view = Matrix.CreateLookAt(new Vector3(50, 10, 0), new Vector3(0, 3, 0), Vector3.UnitY);
            projection = Matrix.CreatePerspectiveFieldOfView(MathHelper.ToRadians(45), GraphicsDevice.Viewport.AspectRatio, 0.1f, 500f);

            // weight: 50 * 9.81 = 490.5
            // LinMom chg = -8.175049 (lin/ang damping = 0) /dt = -490.4983 or .5023

            demo = new Entity(new CapsuleShape(4, 1), 50)
                {
                    Position = new BEPUutilities.Vector3(0, 500, 0),
                    LinearDamping = 0f,
                    AngularDamping = 0f,
                    PositionUpdateMode = PositionUpdateMode.Continuous
                };

            Space.Add(demo);
            ModelDrawer.Add(demo);

            base.Initialize();
        }

        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);
            osdfont = Content.Load<SpriteFont>("osd");

            IsFixedTimeStep = true;

            ground = new Entity(new BoxShape(50000, 1, 50000)) { Position = BEPUutilities.Vector3.Zero };
            ModelDrawer.Add(ground);
            Space.Add(ground);

            GC.Collect();
        }

        protected override void UnloadContent()
        {
        }

        protected override void Update(GameTime gameTime)
        {
            PreviousKeyboardInput = KeyboardInput;
            KeyboardInput = Keyboard.GetState();
            PreviousMouseInput = MouseInput;
            MouseInput = Mouse.GetState();

            var dt = (float)gameTime.ElapsedGameTime.TotalSeconds;

            if (KeyboardInput.IsKeyDown(Keys.Escape))
            {
                Exit();
            }

            if (KeyboardInput.IsKeyDown(Keys.P) && PreviousKeyboardInput.IsKeyUp(Keys.P))
            {
                System.Diagnostics.Debug.Print("PAUSE!!");
                pause = !pause;
            }

            if (pause == true)
            {
                return;
            }

            if (KeyboardInput.IsKeyDown(Keys.Up))
            {
                demo.LinearMomentum += new BEPUutilities.Vector3(0, 800 * dt,0);

            }

            

            if (KeyboardInput.IsKeyDown(Keys.L) && PreviousKeyboardInput.IsKeyUp(Keys.L))
            {
                levitateMode = !levitateMode;
            }

            view = Matrix.CreateLookAt(new Vector3(80, 3, 0), MathConverter.Convert(demo.Position), Vector3.UnitY);

            PreviousLinearMomentum = demo.LinearMomentum;

            
            if (levitateMode == true)
            {
                // get current momentum + way to get zero momentum
                float targetLM = -demo.LinearMomentum.Y;

                System.Diagnostics.Debug.Print(string.Format("tLM={0} LM={1} aLM={2}", targetLM, demo.LinearMomentum.Y, targetLM * dt));
                //demo.LinearMomentum = BEPUutilities.Vector3.Zero; // 

                //demo.LinearMomentum -= demo.LinearMomentum; // +demo.LinearMomentum * dt; //+= new BEPUutilities.Vector3(0, targetLM, 0);
                // pb: LM is null BEFORE space update which will add the step LM

                // 3 targetLM = ~ -3 m/s linvel, -163 LM
                // even -LM => tLM=8,018867 LM=-8,018867 aLM=0,133648   == LM not null !, LV is neg
                // if before space update => tLM=8,018867 LM=-8,018867 aLM=0,133648 = same result !

                //demo.LinearMomentum += new BEPUutilities.Vector3(0, 490.5f * dt, 0);
            }

            Space.Update();
            ModelDrawer.Update();

            // with or without LM = 0, pos still decrease without any velocity AT ALL !
            if (levitateMode == true)
            {
                demo.LinearMomentum -= demo.LinearMomentum;
                System.Diagnostics.Debug.Print(string.Format("LM2={0}", demo.LinearMomentum.Y));
            }

            LinearMomentum = demo.LinearMomentum;

            System.Diagnostics.Debug.Print(string.Format("LM={0} PLM={1}", LinearMomentum, PreviousLinearMomentum));

            PreviousKeyboardInput = KeyboardInput;
            PreviousMouseInput = MouseInput;

            base.Update(gameTime);
        }

        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.DarkSlateBlue);
            GraphicsDevice.DepthStencilState = new DepthStencilState() { DepthBufferEnable = true };

            string text = string.Format("Pos={0} LinVel={1} LinMom={2}",demo.Position, demo.LinearVelocity, demo.LinearMomentum );
            text += string.Format("   LinMomDiff={0}", (LinearMomentum - PreviousLinearMomentum) / (float)gameTime.ElapsedGameTime.TotalSeconds);

            Console.WriteLine(text);
            ModelDrawer.Draw(MathConverter.Convert(view), MathConverter.Convert(projection));

            spriteBatch.Begin();
            spriteBatch.DrawString(osdfont, text, new Vector2(10, 10), Color.White);
            spriteBatch.End();

            base.Draw(gameTime);
        }
    }
}